\datethis
\def\em{\bf}
\font\ff=cmr12
\ff
This program converts Windows TrueType Fonts into Type1 font for use in \TeX.
It's originally writen by He Qiang\footnote\dag{\em obase@@163.net}. 
Modified by Wang Yin\footnote\ddag{\em wang-y01@@mails.tsinghua.edu.cn} 
into a \.{CWEB} file and changed some portions of the code.

@* The gbkfonts Program. 

The program is making use of several other program from the FreeType Project 
({\em http://www.freetype.org}) such as \.{ttf2pfb, t1asm, type1afm, afm2tfm}.
The following is the overall structure of the program text.

@c
@<Included files@>@/
@<Global variables@>@/
@<Utility functions@>@/
@<The |main| program@>@/

@ @<Include...@>=
#include <stdio.h>
#include <process.h>
#include <string.h>
#include <direct.h>
#include <stdarg.h>
#include <windows.h>
#include <winbase.h>
#include <signal.h>

@ @<Global...@>=
static char gb[128], gbk[128], compact[128];
static char gbkface[128], slface[128];
static char name[128], slname[128];
static char ttf_file[128], ttf_basename[128], fontname[128];
static char cmd[128], cmd1[128], plane[8];
static int  bt, bc, bgbk, bv, bf;


@ Utilities for report errors.
@<Utility...@>=
void fatal(char *str)
{
	fprintf(stderr, str);
	exit(-1);
}

void show(char *str)
{
	if (bv) {
		fprintf(stderr, str);
		fprintf(stderr, "\n");
	}
}

@ some variables
@<variable...@>=
	FILE * fin;
	char buf[128], *ptr;

@ call ftdump to get the TTF font eid. 
@ @<get eid...@>=
sprintf(buf, "ftdump %s", ttf_file);
if ((fin = _popen(buf, "r")) == NULL) 
	fatal ("can't find ftpdump program\n");

while (!feof(fin))
{
	if (fgets(buf,128,fin) == 0) 
		fatal("Error: file encoding not supported.");

	if (ptr = strstr(buf, "Windows Unicode")) {
		eid = 1; break; 
	}

	if (ptr = strstr(buf, "Windows GB2312")) {
		eid = 3; break; 
	}
}

fclose(fin);

@ open the gbkfonts.ini file and get the name of font, the slant name
  of the font, etc. The font uid can be set by the user.

@<get uid...@>=
if ((fin = fopen("gbkfonts.ini","rt")) == NULL)
fatal("can't open gbkfonts.ini\n");

while (!feof(fin))
{   
	n = fscanf(fin,"%s %s %s %d", s_face, slface, fontname, &fontuid);
	if (n!=4) continue;
	if (strcmp(gbkface,s_face)==0) break;
	fontuid = 0;
}
fclose(fin);

if (fontuid == 0) fatal("Can find font UID in gbkfonts.ini");

@ Create fd files such as C19song.fd for CJK package. 
I omit the $\backslash$CJKbold option in the fd file, 
since most user don't use them.

@<create fd...@>= 
	/* write fd file for latex */
	sprintf(cmd, "tex/latex/CJK/GB/c19%s.fd", gbkface);
	fd_file = fopen (cmd, "wt");
	fprintf(fd_file, "%% This is the file c19%s.fd of the CJK package\n", gbkface);
	fprintf(fd_file, "%% for using Asian logographs (Chinese/Japanese/Korean) with LaTeX2e\n");
	fprintf(fd_file, "%% created by Werner Lemberg <wl@@gnu.org>\n");
	fprintf(fd_file, "%% Version 4.3.0 (20-Jun-1999)\n\n");
	fprintf(fd_file, "\\def\\fileversion{4.3.0}\n");
	fprintf(fd_file, "\\def\\filedate{1999/06/20}\n\n");
	fprintf(fd_file, "%% \\ProvidesFile{c19%s.fd}[\\filedate\\space\\fileversion]\n", gbkface);
	fprintf(fd_file, "%% Chinese characters (extension of GB 2312)\n");
	fprintf(fd_file, "%% character set: GBK\n");
	fprintf(fd_file, "%% font encoding: CJK (extended)\n\n");
	fprintf(fd_file, "\\DeclareFontFamily{C19}{%s}{}\n", gbkface);
	fprintf(fd_file, "\\DeclareFontShape{C19}{%s}{m}{n}{<-> CJK * %s%s}{}\n", gbkface, gb, gbkface);
	fprintf(fd_file, "\\DeclareFontShape{C19}{%s}{bx}{n}{<-> CJKb * %s%s}{}\n", gbkface, gb, gbkface);
	fprintf(fd_file, 
		"\\DeclareFontShape{C19}{%s}{m}{it}{<-> CJK * %s%ssl}{}\n", 
		gbkface, gb, gbkface);
	fprintf(fd_file, 
		"\\DeclareFontShape{C19}{%s}{bx}{it}{<-> CJKb * %s%ssl}{}\n", 
		gbkface, gb, gbkface);
	fprintf(fd_file, 
		"\\DeclareFontShape{C19}{%s}{m}{sl}{<-> CJK * %s%ssl}{}\n",   
		gbkface, gb, gbkface);
	fprintf(fd_file, 
		"\\DeclareFontShape{C19}{%s}{bx}{sl}{<-> CJKb * %s%ssl}{}\n", 
		gbkface, gb, gbkface);
	fprintf(fd_file, "\n\\endinput\n");
	fclose(fd_file);


@ This function get the file basename's pointer.
@<Utility...@>=
char *trimpath(char* str)
{	
	
	char* ptr;
	ptr = strrchr(str, '\\');
	if (ptr == 0) ptr = strrchr(str, '/');
	if (ptr == 0) return str; else return ptr+1;
}

@ @<variables us...@>=
int  i, eid, fontuid, n, nn, total_char, offset;
FILE *pfbmapfile, *ttfmapfile, *fout, *dvipdfm_mapfile, *fd_file;
char s_face[128];

@ set the names. for example, if you use\hfil\break
\tt gbkfonts -c -gbk c:/winnt/fonts/simsun.ttf song \rm

you get the names: name = ``gbksong??'' , slname = ``gbksongsl??''

@<set appro...@>=
	if (bc)
		sprintf(plane, "%02d", nn);
	else
		sprintf(plane, "%02x", nn);
	
	sprintf(name,  "%s%s%s", gb, gbkface, plane);
	sprintf(slname,"%s%s%s", gb, slface, plane);

@ Write configuration files for several programs.

The {\tt dvipdfmx} program from the dvipdfm-cjk project use cid-x.map to 
find TTF Chinese fonts. 

Pdftex use pdftex.cfg to find the map files. The map files direct it to the TTF
fonts. They are the same as used by {\tt dvips} when {\tt dvips} use PK fonts.

The {\tt CJK} package find it's font file name from the .fd files.

@<write config...@>=
	/* write cid-x for dvipdfm */
	fout = fopen ("cid-x.map", "a");
	fprintf(fout, "gb%s@@UGBK@@             UniGB-UCS2-H    :0:%s\n", gbkface, ttf_basename);

	fprintf(fout, "gb%ssl@@UGBK@@             UniGB-UCS2-H    :0:%s\n", gbkface, ttf_basename);
	fclose (fout);

	/* write pdftex.cfg */
	fout = fopen ("pdftex.cfg", "a");
	fprintf(fout, "map +%s%s.map\n", gb, gbkface);
	fclose (fout);

	/* write map files for dvips */
	fout = fopen ("ttfonts.map", "a");
	fprintf(fout, "%s%s@@UGBK@@    %s  Pid = 3 Eid = %d\n", gb, gbkface, ttf_basename, eid);
	fprintf(fout, "%s%s@@UGBK@@    %s  Slant = 0.167 Pid = 3 Eid = %d\n", 
			gb, slface, ttf_basename, eid);
	fclose (fout);

	@<create fd file for LaTeX CJK@>@/

@ @<Utility...@>=
void showhelp()
{
	puts("Usage:   gbkfonts [-c] [-v] [-f] [-t] [-gbk] ttf_file gbkface");
	puts("\t      -c: create CJK compact font");
	puts("\t      -v: verbose, show the commands");
	puts("\t      -f: only create config and map files");
	puts("\t      -t: create traditional font");
	puts("\t    -gbk: add this if your ttf is GBK font");
	puts("\tttf_file: TTF file name");
	puts(" gbkface: song/hei/kai/li/fs, etc. see gbkfonts.ini");
	puts("");
	puts("Make pfb, afm, tfm and enc files from ttf_file,");
	puts("also make all map files needed.");
	puts("The font definitions are in gbkfonts.ini.");
	puts("");
	puts("Example: gbkfonts -c -gbk c:\\windows\\fonts\\simsun.ttf song");
	puts("");
	puts("Then copy folder \"dvips\", \"tex\", \"dvipdfm\", \
and \"fonts\" to TEXMF-LOCAL,");
	puts("append file cjk.map to c:\\TeX\\texmf\\dvips\\base\\cjk.map,");
	puts("append file cid-x.map to c:\\TeX\\texmf\\dvipdfm\\config\\\
cid-x.map,");
}

@ @<The |main| progra...@>=
int main(int argc, char* argv[])
{
	@<variables used in |main| @>;
	@<register signal handler@>;
	@<initialize strings@>;
	@<parse arguments, test font file existence@>;
	@<get eid : encode id from the TTF font@>;
	@<get uid information from the gbkfonts.ini file@>;
	@<decide total number of characters in the font store it in\
		variable |total_char|@>;
	@<create directory to store fonts and other files@>;
	@<open map files for write@>;

	//fclose(stdout); /* close the screen out put from the called programs */
	//fopen("nul", "w");

	fprintf(stderr, "Processing %s%s\n", gb, gbkface);@/

	for (n = 1; n <= total_char; n++) {
		
		fprintf(stderr, "[%d]", n); 
		@<decide plane number, store it in |nn|@>;
		@<set appropriate names {\tt name}, {\sl slname} $\ldots$ for the font@>;

		if (bf) goto skipfontmake;
		@<make the fonts: {\sl pfb, afm, tfm} @>;
		@<move file to the destination@>;
skipfontmake:
		@<add line to map files@>;
	}
	@<close mapfiles@>;
	@<write config files@>;
	return 0;
}

@ @<initializ...@>=
strcpy(gb       ,"gb");
strcpy(gbk      ,"");
strcpy(compact  ,"");
strcpy(ttf_file ,"");
strcpy(gbkface     ,"");


@ @<parse...@>=
if (argc<2 || argc>6)
{   
	showhelp();
	return -1;
}

for (i = 1; i < argc; i++)
{
	if (strcmp(argv[i], "-c") == 0) { 
		strcpy(compact,"-c");
		bc = 1;
	}@+ else if (strcmp(argv[i], "-gbk") == 0) { 
		strcpy(gbk,"-gbk");
		bgbk = 1;
		strcpy(gb, "gbk");
	}@+ else if (strcmp(argv[i], "-t") == 0) { 
		strcpy(gb,"gt");
		bt = 1;
	}@+ else if (strcmp(argv[i], "-v") == 0) bv = 1;
	else if (strcmp(argv[i], "-f") == 0) bf = 1; 
	else if (ttf_file[0]==0) strcpy(ttf_file,argv[i]);
	else if (gbkface[0]==0) strcpy(gbkface,argv[i]);
	else {   
		showhelp();
		return -1;
	}
}
fprintf(stderr, "%s\n", ttf_file);
if (_access(ttf_file, 0) == -1) 
	fatal ("Can't open font file");
strcpy (ttf_basename, trimpath (ttf_file));

@ @<create dire...@>=
_mkdir("fonts");
_mkdir("fonts/afm");
_mkdir("fonts/tfm");
_mkdir("fonts/type1");
_mkdir("fonts/afm/chinese");
_mkdir("fonts/tfm/chinese");
_mkdir("fonts/type1/chinese");

_mkdir("dvips");
_mkdir("dvips/chinese");

_mkdir("pdftex");
_mkdir("pdftex/config");

_mkdir("dvipdfm");
_mkdir("dvipdfm/config");

_mkdir("tex");
_mkdir("tex/latex");
_mkdir("tex/latex/CJK");
_mkdir("tex/latex/CJK/GB");

sprintf(cmd,"dvips/chinese/%s%s", gb,gbkface);
_mkdir(cmd);

sprintf(cmd,"fonts/afm/chinese/%s%s", gb,gbkface);
_mkdir(cmd);

sprintf(cmd,"fonts/tfm/chinese/%s%s", gb,gbkface);
_mkdir(cmd);

sprintf(cmd,"fonts/type1/chinese/%s%s", gb,gbkface);
_mkdir(cmd);

@ @<decide tot...@>=
if ( bc &&  bgbk) total_char =  94;
else if ( bc && !bgbk) total_char =  35;
else if (!bc &&  bgbk) total_char =  126;
else total_char =  81;

@ @<open map...@>=
pfbmapfile = fopen ("cjk.map", "at");
sprintf(cmd,"pdftex/config/%s%s.map", gb, gbkface);
ttfmapfile = fopen (cmd, "wt");
sprintf(cmd,"dvipdfm/config/%s%s.map", gb, gbkface);
dvipdfm_mapfile = fopen (cmd, "wt");

@ @<decide plane...@>=
if (bc) offset = 0;
else if (bgbk)  offset = 0x80;
else if (n<=9)  offset = 0xa0;
else offset = 0xa6;
nn = n + offset;

@ @<make the fon...@>=
/* Make .enc and .pfb from .ttf */
sprintf(cmd,"ttf2pfb -a -pid 3 -eid %d %s -force UGBK \
	-plane %d %s -uid %d -f %s %s -o %s.ps",
eid, gbk, nn, compact, fontuid, fontname, ttf_file, name);
show(cmd);
system(cmd);

/* assemble row fonts into type1 pfb */
sprintf(cmd,"t1asm -b %s.ps > %s.pfb", name, name);
show(cmd);
system(cmd);

/* delete the ps file */
sprintf(cmd, "%s.ps", name);
_unlink(cmd);

/* Make .afm from .pfb */
sprintf(cmd, "type1afm %s.pfb", name);
system (cmd);

/* appendline to cjk.Fontmap */
fout = fopen("cjk.Fontmap", "a");
fprintf(fout ,"/%s%s (%s.pfb) ;\n", fontname, plane, name);
fclose(fout);

/* Make .tfm from .afm */
sprintf(cmd,"afm2tfm %s.afm", name);
show(cmd);
system(cmd);

/* make slant tfm font */
sprintf(cmd,"afm2tfm %s.afm -s .167 %s.tfm", name, slname);
show(cmd);
system(cmd);


@ @<move...@>=
/*afm*/
sprintf(cmd, "%s.afm", name);
sprintf(cmd1, "fonts/afm/chinese/%s%s/%s.afm", gb, gbkface, name);
MoveFileEx(cmd, cmd1, MOVEFILE_REPLACE_EXISTING);

/*tfm*/
sprintf(cmd, "%s.tfm", name);
sprintf(cmd1, "fonts/tfm/chinese/%s%s/%s.tfm", gb, gbkface, name);
MoveFileEx(cmd, cmd1, MOVEFILE_REPLACE_EXISTING);

/* slant tfm */
sprintf(cmd, "%s.tfm", slname);
sprintf(cmd1, "fonts/tfm/chinese/%s%s/%s.tfm", gb, gbkface, slname);
MoveFileEx(cmd, cmd1, MOVEFILE_REPLACE_EXISTING);

/*pfb*/
sprintf(cmd, "%s.pfb", name);
sprintf(cmd1, "fonts/type1/chinese/%s%s/%s.pfb", gb, gbkface, name);
MoveFileEx(cmd, cmd1, MOVEFILE_REPLACE_EXISTING);

/*enc*/
sprintf(cmd, "%s.enc", name);
sprintf(cmd1, "dvips/chinese/%s%s/%s.enc", gb, gbkface, name);
MoveFileEx(cmd, cmd1, MOVEFILE_REPLACE_EXISTING);

@ @<add line...@>=
/* add line to map file for dvips use ttf fonts */
fprintf(ttfmapfile,"%s <%s.enc <%s\n", name, name, ttf_basename);

fprintf(ttfmapfile,"%s <%s.enc <%s\n", slname, name, ttf_basename);
fprintf(dvipdfm_mapfile,"%s none %s\n", name, name);

fprintf(dvipdfm_mapfile,"%s none %s -s .167\n", slname, name);

/* map file for dvips use pfb fonts */
fprintf(pfbmapfile,"%s %s%s <%s.pfb\n", name, fontname, plane, name);

fprintf(pfbmapfile,"%s %s%s \" .167 SlantFont \" < %s.pfb\n", 
		slname, fontname, plane, name);

@ @<close map...@>=
	fclose(ttfmapfile);
	fclose(pfbmapfile);
	fclose(dvipdfm_mapfile);
	_unlink("cjk.Fontmap");

@ @<register...@>=
signal(SIGINT, sig_int);

@ @<Utili...@>=
void sig_int(int sig)
{
	_unlink("cjk.Fontmap");
	fprintf(stderr, "\n\n! Interrupted by user\n");
	exit(2);
}

